import os
import textwrap
import warnings
from numbers import Number
from pathlib import Path
from typing import List, Literal, Optional, Tuple, Union
from warnings import warn

import numpy as np
import pandas as pd
import seaborn as sns
from anndata import AnnData
from matplotlib import pyplot as plt
from matplotlib.axes._axes import Axes
from matplotlib.figure import Figure
from scipy.stats import pearsonr, spearmanr, zscore
from sklearn.preprocessing import MinMaxScaler
from tqdm import tqdm

from insitupy._constants import DEFAULT_CATEGORICAL_CMAP, _init_mpl_fontsize
from insitupy._core._checks import check_raw, has_valid_labels
from insitupy.io.plots import save_and_show_figure
from insitupy.utils._regression import smooth_fit
from insitupy.utils.utils import (convert_to_list, get_nrows_maxcols,
                                  remove_empty_subplots)


def expr_along_obs_val(
    adata: AnnData,
    keys: str,
    obs_val: Union[str, Tuple[str, str]],
    groupby: Optional[str] = None,
    splitby: str = None,
    hue: str = None,
    method: Literal["lowess", "loess"] = 'loess',
    stderr: bool = False,
    loess_bootstrap: bool = True,
    n_bootstraps_iterations: int = 100,
    xmin=None,
    xmax=None,
    cmap="tab10",
    linewidth=8,
    extra_cats=None,
    normalize=False,
    nsteps=100,
    show_progress=False,
    use_raw=False,
    max_cols=4,
    xlabel=None,
    ylabel=None,
    vline=None,
    hline=None,
    vlinewidth=4,
    custom_titles=None,
    legend_fontsize=24,
    plot_legend=True,
    xlabel_fontsize=28,
    ylabel_fontsize=28,
    title_fontsize=20,
    tick_fontsize=24,
    figsize=(8,6),
    savepath: Optional[os.PathLike] = None,
    save_only: bool = False,
    show: bool = True,
    return_data: bool = False,
    fig: Optional[Figure] = None,
    axis: Optional[Axes] = None,
    dpi_save: int = 300,
    smooth=True,
    **kwargs
    ):
    """
    Plot gene expression values along a specified observation category.

    Args:
        adata (AnnData): Annotated data matrix.
        keys (str): Keys for the gene expression values to be plotted.
        obs_val (Union[str, Tuple[str, str]]): Observation category to be plotted on the x-axis.
            Can be a string representing a column in `adata.obs` or a tuple (obsm_key, obsm_col)
            where `obsm_key` is a key in `adata.obsm` and `obsm_col` is a column in the corresponding DataFrame.
        groupby (Optional[str]): Observation category to group by.
        splitby (str, optional): Observation category to split by.
        hue (str, optional): Observation category to color by.
        method (Literal["lowess", "loess"], optional): Smoothing method to use. Defaults to 'loess'.
        stderr (bool, optional): Whether to plot standard error. Defaults to False.
        loess_bootstrap (bool, optional): Whether to use bootstrap for loess smoothing. Defaults to True.
        n_bootstraps_iterations (int, optional): Number of bootstrap iterations for loess smoothing. Defaults to 100.
        xmin (optional): Minimum x value for plotting.
        xmax (optional): Maximum x value for plotting.
        cmap (str, optional): Colormap to use for plotting. Defaults to "tab10".
        linewidth (int, optional): Line width for plotting. Defaults to 8.
        extra_cats (optional): Additional observation categories to include in the plot.
        normalize (bool, optional): Whether to normalize the expression values. Defaults to False.
        nsteps (int, optional): Number of steps for smoothing. Defaults to 100.
        show_progress (bool, optional): Whether to show progress bar. Defaults to False.
        use_raw (bool, optional): Whether to use raw data. Defaults to False.
        max_cols (int, optional): Maximum number of columns for subplots. Defaults to 4.
        xlabel (optional): Label for the x-axis.
        ylabel (optional): Label for the y-axis.
        vline (optional): Vertical lines to add to the plot.
        hline (optional): Horizontal lines to add to the plot.
        vlinewidth (int, optional): Line width for vertical lines. Defaults to 4.
        custom_titles (optional): Custom titles for the plots.
        legend_fontsize (int, optional): Font size for the legend. Defaults to 24.
        plot_legend (bool, optional): Whether to plot the legend. Defaults to True.
        xlabel_fontsize (int, optional): Font size for the x-axis label. Defaults to 28.
        ylabel_fontsize (int, optional): Font size for the y-axis label. Defaults to 28.
        title_fontsize (int, optional): Font size for the plot titles. Defaults to 20.
        tick_fontsize (int, optional): Font size for the axis ticks. Defaults to 24.
        figsize (tuple, optional): Figure size. Defaults to (8, 6).
        savepath (optional): Path to save the plot.
        save_only (bool, optional): Whether to only save the plot without showing. Defaults to False.
        show (bool, optional): Whether to show the plot. Defaults to True.
        axis (optional): Axis to plot on.
        return_data (bool, optional): Whether to return the data instead of plotting. Defaults to False.
        fig (optional): Figure to plot on.
        dpi_save (int, optional): DPI for saving the plot. Defaults to 300.
        smooth (bool, optional): Whether to apply smoothing. Defaults to True.
        **kwargs: Additional arguments for smoothing.

    Returns:
        Union[DataFrame, Tuple[Figure, Axes]]:
            If return_data is True, returns a DataFrame with the smoothed data.
            Otherwise, returns the figure and axes of the plot.
    """

    # check type of input
    if isinstance(keys, dict):
        if custom_titles is not None:
            print("Attention: `custom_titles` was not None and `keys` was dictionary. Titles were retrieved from dictionary.")
        custom_titles = list(keys.keys())
        keys = list(keys.values())

    # make inputs to lists
    keys = [keys] if isinstance(keys, str) else list(keys)


    adata_obs = adata.obs.copy()
    if isinstance(obs_val, tuple):
        print("Retrieve `obs_val` from .obsm.")
        obsm_key = obs_val[0]
        obsm_col = obs_val[1]
        obs_val = f"distance_from_{obsm_col}"
        adata_obs[obs_val] = adata.obsm[obsm_key][obsm_col]

    # remove NaNs `obs_val` column
    not_na_and_not_zero_mask = adata_obs[obs_val].notna() & adata_obs[obs_val] > 0
    adata_obs = adata_obs[not_na_and_not_zero_mask]

    # check whether to plot raw data
    X, var, var_names = check_raw(adata, use_raw=use_raw)

    # remove rows from X which were NaN above
    X = X[not_na_and_not_zero_mask]

    if hue is not None:
        hue_cats = list(adata_obs[hue].unique())
        cmap_colors = plt.get_cmap(cmap)
        color_dict = {a: cmap_colors(i) for i, a in enumerate(hue_cats)}

        if extra_cats is None:
            extra_cats = [hue]
        else:
            extra_cats.append(hue)

    #if show:
    if not return_data:
        # prepare plotting
        if axis is None:
            n_plots, n_rows, max_cols = get_nrows_maxcols(len(keys), max_cols)
            fig, axs = plt.subplots(n_rows,max_cols, figsize=(figsize[0]*max_cols, figsize[1]*n_rows))

        else:
            axs = axis
            #fig = None
            n_plots = 1
            show = False # otherwise plotting into given axes wouldn't work

        if n_plots > 1:
            axs = axs.ravel()
        else:
            axs = [axs]

    data_collection = {}
    for i, key in (enumerate(tqdm(keys)) if show_progress else enumerate(keys)):
        # check if the keys are also grouped
        keys_grouped = isinstance(key, list)

        if groupby is not None:
            # select data per group
            groups = adata_obs[groupby].unique()
        else:
            groups = [None]

        added_to_legend = []

        group_collection = {}
        for group in groups:
            #partial = extract_groups(adata, groupby=groupby, groups=group)

            if group is not None:
                group_mask = adata_obs[groupby] == group
                group_obs = adata_obs.loc[group_mask, :].copy()
            else:
                group_mask = [True] * len(adata_obs)
                group_obs = adata_obs

            if hue is not None:
                _hue = adata_obs.loc[group_mask, hue][0]

            # hue_data = adata_obs.loc[group_mask, hue].copy()
            # print(hue_data)

            # select only group values from matrix
            group_X = X[group_mask, :]

            if splitby is None:
                # select x value
                x = group_obs.loc[:, obs_val].values
                # if xmin is None:
                #     xmin = x[x>0].min()
                #     print(xmin, flush=True)

                if keys_grouped:
                    # extract expression values of all keys in the group
                    idx = var.index.get_indexer(key)
                    dd = pd.DataFrame(group_X[:, idx], index=x)

                    if normalize:
                        #dd = dd.apply(minmax_scale, axis=0)
                        dd = dd.apply(zscore, axis=0)

                    dd = dd.reset_index().melt(id_vars="index") # reshape to get long list of x values
                    x = dd["index"].values
                    y = dd["value"].values

                elif key in var_names:
                    # extract expression values as y
                    idx = var.index.get_loc(key)
                    y = group_X[:, idx].copy()

                    if normalize:
                        #y = minmax_scale(y)
                        y = zscore(y)

                elif key in group_obs.columns:
                    y = group_obs.loc[:, key].values.copy()
                else:
                    print("Key '{}' not found.".format(key))
                    break

                if smooth:
                    # do smooth fitting
                    df = smooth_fit(x, y,
                                    xmax=xmin, xmin=xmax,
                                    nsteps=nsteps, method=method,
                                    stderr=stderr, loess_bootstrap=loess_bootstrap,
                                    K=n_bootstraps_iterations,
                                    **kwargs)
                else:
                    # set up dataframe without smooth fitting
                    df = pd.DataFrame({"x": x, "y_pred": y})

                if extra_cats is not None:
                    df = df.join(adata_obs.loc[group_mask, extra_cats].reset_index(drop=True))

            else:
                splits = group_obs[splitby].unique()
                df_collection = {}

                # get min and max values for x values
                x = group_obs[obs_val].values
                xmin = x.min()
                xmax = x.max()

                for split in splits:
                    # extract x values
                    split_mask = group_obs[splitby] == split
                    x = group_obs.loc[split_mask, obs_val].values

                    # extract expression values as y
                    idx = var.index.get_loc(key)
                    y = group_X[split_mask, idx].copy()

                    # do smooth fitting
                    if smooth:
                        df_split = smooth_fit(x, y,
                                xmax=xmin, xmin=xmax,
                                nsteps=nsteps, method=method, stderr=stderr, **kwargs)
                    else:
                        # set up dataframe without smooth fitting
                        df_split = pd.DataFrame({"x": x, "y_pred": y})

                    # collect data
                    df_collection[split] = df_split

                df_collection = pd.concat(df_collection)

                # calculate mean and std
                df = df_collection[['x', 'y_pred']].groupby('x').mean()
                df['std'] = df_collection[['x', 'y_pred']].groupby('x').std()
                df['conf_lower'] = [a-b for a,b in zip(df['y_pred'], df['std'])]
                df['conf_upper'] = [a+b for a,b in zip(df['y_pred'], df['std'])]
                df.reset_index(inplace=True)

            # remove NaNs
            df = df.dropna(how="all", axis=1)

            if return_data:
                group_collection[group] = df
            else:
                # sort by x-value
                df.sort_values('x', inplace=True)

                # plotting
                cols = df.columns
                if 'conf_lower' in cols and 'conf_upper' in cols:
                    axs[i].fill_between(df['x'],
                                    df['conf_lower'],
                                    df['conf_upper'],
                                    alpha = 0.2,
                                    color = 'grey')

                # determine label variable
                if hue is not None:
                    label = _hue if _hue not in added_to_legend else ""
                    color = color_dict[_hue]
                else:
                    label = group
                    color = None

                axs[i].plot(df['x'],
                    df['y_pred'],
                    label=label,
                    color=color,
                    linewidth=linewidth)

                if hue is not None and _hue not in added_to_legend:
                    added_to_legend.append(_hue)

        # optionally add vertical or horizontal lines to plot
        if vline is not None:
            if isinstance(vline, dict):
                linecolors = list(vline.keys())
                vline = list(vline.values())
            else:
                vline = [vline] if isinstance(vline, int) or isinstance(vline, float) else list(vline)
                linecolors = ['k'] * len(vline)

            for c, v in zip(linecolors, vline):
                axs[i].axvline(x=v, ymin=0, ymax=1, c=c, linewidth=vlinewidth, linestyle='dashed')

        if hline is not None:
            if isinstance(hline, dict):
                linecolors = list(hline.keys())
                hline = list(hline.values())
            else:
                hline = [hline] if isinstance(hline, int) or isinstance(hline, float) else list(hline)
                linecolors = ['k'] * len(hline)

            for c, h in zip(linecolors, hline):
                axs[i].axhline(y=h, xmin=0, xmax=1, c=c, linewidth=4, linestyle='dashed')

        if not return_data:
            if xlabel is None:
                xlabel = obs_val
            if ylabel is None:
                ylabel = "Gene expression"

            axs[i].set_xlabel(xlabel, fontsize=xlabel_fontsize)
            axs[i].set_ylabel(ylabel, fontsize=ylabel_fontsize)
            axs[i].tick_params(axis='both', which='major', labelsize=tick_fontsize)
            #axs[i].set_xlim(0, 1)
            #axs[i].xaxis.set_major_locator(ticker.FixedLocator([0.1, 0.9]))

            if custom_titles is None:
                axs[i].set_title(key, fontsize=title_fontsize)
            else:
                assert len(custom_titles) == len(keys), "List of title values has not the same length as list of keys."
                axs[i].set_title(str(custom_titles[i]), fontsize=title_fontsize)

            if plot_legend:
                if has_valid_labels(axs[i]):
                    axs[i].legend(fontsize=legend_fontsize,
                    loc='best'
                    )
            else:
                # first check if there are valid labels in the axis to circumvent warning
                if has_valid_labels(axs[i]):
                    axs[i].legend().remove()

        if return_data:
            if len(group_collection) > 0:
                # collect data
                group_collection = pd.concat(group_collection)
                data_collection[key] = group_collection
            else:
                pass

    if return_data:
        # close plot
        plt.close()

        # return data
        data_collection = pd.concat(data_collection)
        data_collection.index.names = ['key', groupby, None]
        return data_collection

    else:
        if n_plots > 1:

            # check if there are empty plots remaining
            while i < n_rows * max_cols - 1:
                i+=1
                # remove empty plots
                axs[i].set_axis_off()
        if show:
            #fig.tight_layout()
            save_and_show_figure(savepath=savepath, fig=fig, save_only=save_only, dpi_save=dpi_save, tight=True)
        else:
            return fig, axs


def cell_expression_along_axis(
    adata,
    axis,
    genes: List[str],
    cell_type_column,
    cell_type,
    xlim: Tuple[Union[int, float], Union[int, float]] = (0, np.inf),
    min_expression: Union[int, float] = 0,
    xlabel: Optional[str] = None,
    fit_reg: bool = False,
    kde: bool = False,
    maxcols: bool = 4,
    savepath: Union[str, os.PathLike, Path] = None,
    save_only: bool = False,
    dpi_save: int = 300,
    fig_height: Number = 4,
    fig_marginal_ratio: Number = 0.15,
    scatter_size: Number = 1,
    wspace: Number = 0.15,
    hspace: Number = 0.25,
    font_scale_factor: Number = 1
):

    """
    Plot gene expression along a specified axis for a given cell type.

    Args:
        adata: AnnData object containing the single-cell data.
        axis: Observation value to plot along the x-axis.
        genes (List[str]): List of genes to plot.
        cell_type_column: Column name in `adata.obs` that contains cell type information.
        cell_type: Specific cell type to filter the data.
        xlim (Tuple[Union[int, float], Union[int, float]], optional): Limits for the x-axis. Defaults to (0, np.inf).
        min_expression (Union[int, float], optional): Minimum expression level to include in the plot. Defaults to 0.
        xlabel (Optional[str], optional): Label for the x-axis. Defaults to None.
        fit_reg (bool, optional): Whether to fit a regression line. Defaults to False.
        lowess (bool, optional): Whether to use LOWESS for regression. Defaults to False.
        robust (bool, optional): Whether to use a robust regression. Defaults to False.
        fig_height (Number, optional): Height of the figure. Defaults to 4.
        fig_marginal_ratio (Number, optional): Ratio of the marginal plot height to the main plot height. Defaults to 0.2.
        scatter_size (Number, optional): Size of the scatter plot points. Defaults to 1.

    Returns:
        None: Displays the plot.
    """
    # reset matplotlib settings
    _init_mpl_fontsize(scale_factor=font_scale_factor)

    # make sure genes is a list
    genes = convert_to_list(genes)

    # select the data for plotting
    data_for_one_celltype = _select_data(
        adata=adata,
        obs_val=axis,
        cell_type_column=cell_type_column,
        cell_type=cell_type,
        genes=genes,
        min_expression=min_expression,
        xlim=xlim,
    )

    # create xlabel string
    if xlabel is None:
        xlabel_str = " ".join(convert_to_list(axis))
        xlabel_str = textwrap.fill(xlabel_str, width=15)
    else:
        xlabel_str = xlabel

    # Prepare a figure with subplots
    num_genes = len(genes)
    num_rows = (num_genes + maxcols - 1) // maxcols
    marg_height = fig_height * fig_marginal_ratio
    fig, axes = plt.subplots(num_rows + 1, maxcols * 2,
                             figsize=(fig_height * (1-hspace) * maxcols + marg_height,
                                      fig_height * (1-wspace) * num_rows + marg_height),
                             sharey='row', sharex='col',
                             gridspec_kw={'height_ratios': [marg_height] + [fig_height]*num_rows,
                                          'width_ratios': [fig_height, marg_height]*maxcols
                                          }
                             )

    # Adjust the space between subplots
    plt.subplots_adjust(wspace=wspace, hspace=hspace)

    for i, gene in enumerate(genes):
        row = i // maxcols + 1
        col = i % maxcols * 2

        if row == 1:
            # Histogram for the x-axis density
            sns.kdeplot(data=data_for_one_celltype, x="axis", ax=axes[0, col],
                        color='darkgray', fill=True)

            # remove the empty axes in between the plots
            axes[0, col+1].remove()

            # remove values axis from histogram
            axes[0, col].get_yaxis().set_visible(False)
            axes[0, col].spines['left'].set_visible(False)
            axes[0, col].spines['right'].set_visible(False)
            axes[0, col].spines['top'].set_visible(False)

        # set xlabel in the last row
        if row == num_rows:
            axes[row, col].set_xlabel(xlabel_str)

        # set ylabel in the first column
        if col == 0 and row != 0:
            axes[row, col].set_ylabel("Gene expression")

        # select data for current gene
        data_for_one_gene = data_for_one_celltype[[gene]].copy()

        if kde:
            # KDE plot
            sns.kdeplot(data=data_for_one_gene.reset_index("axis"),
                        x="axis", y=gene,
                        ax=axes[row, col],
                        fill=True, cmap="Reds", levels=8)

        # Scatter plot
        axis_values = data_for_one_gene.index.get_level_values('axis').values
        expr_values = data_for_one_gene[gene].values

        # drop NaNs
        not_nan = ~np.isnan(expr_values)
        axis_values = axis_values[not_nan]
        expr_values = expr_values[not_nan]

        axes[row, col].scatter(
            x=axis_values,
            y=expr_values,
            alpha=0.5, color='k', s=2
            )

        if fit_reg:
            if len(axis_values) > 1:
                try:
                    # perform loess regression for the second half of the plot
                    res = smooth_fit(
                    xs=axis_values,
                    ys=expr_values, # make sure there are no NaN in the data
                    loess_bootstrap=False, nsteps=100
                    )
                except ValueError as e:
                    print(f"A ValueError occurred during loess regression: {e}")
                    res = None
            else:
                print(f"Only one datapoint left for gene {gene} after filtering. Skipped LOESS regression.")
                res = None

            if res is not None:
                axes[row, col].plot(res["x"], res["y_pred"],
                            color='royalblue', linewidth=3,
                            #label=reg_label
                            )
                axes[row, col].fill_between(res["x"],
                                    res["conf_lower"],
                                    res["conf_upper"],
                                    color='royalblue',
                                    alpha=0.2,
                                    #label='95% CI of Loess Regression'
                                    )


        # Histogram for the gene expression
        sns.kdeplot(
            data=data_for_one_gene, y=gene, ax=axes[row, col+1], color='darkgray', fill=True
        )

        # remove values axis from histogram
        axes[row, col+1].get_xaxis().set_visible(False)
        axes[row, col+1].spines['bottom'].set_visible(False)
        axes[row, col+1].spines['right'].set_visible(False)
        axes[row, col+1].spines['top'].set_visible(False)

        # Set labels
        #axes[row, col].set_ylabel(f"{gene} in '{cell_type}'")
        #axes[row, col].set_ylabel(f"{gene}")
        axes[row, col].set_title(f"{gene}", y=1.0#pad=-2
                                 )

    plt.suptitle(f"Gene expression in '{cell_type}'")

    # Turn off empty subplots
    total_plots = (num_rows + 1) * maxcols * 2
    for i in range(len(genes), maxcols * num_rows):
        row = i // maxcols + 1
        col = i % maxcols * 2
        axes[row, col].set_axis_off()
        axes[row, col + 1].set_axis_off()

        if row == 1:
            axes[0, col].set_axis_off()
            axes[0, col+1].set_axis_off()

        # add the xlabel to the plot above
        axes[row-1, col].set_xlabel(xlabel_str)

    save_and_show_figure(savepath=savepath, fig=fig, save_only=save_only, dpi_save=dpi_save, tight=False)


def _select_data(
    adata,
    obs_val,
    genes: List[str],
    cell_type_column,
    cell_type,
    xlim: Tuple[Union[int, float], Union[int, float]] = (0, np.inf),
    min_expression: Number = None,
    sort: bool = True,
    minmax_scale: bool = True,
    verbose: bool = True
):
    # make sure genes is a list
    genes = convert_to_list(genes)

    # Check type of obs_val
    adata_obs = adata.obs.copy()
    if isinstance(obs_val, tuple):
        print("Retrieve `obs_val` from .obsm.") if verbose else None
        obsm_key = obs_val[0]
        obsm_col = obs_val[1]
        #obs_val = f"distance_from_{obsm_col}"
        adata_obs["axis"] = adata.obsm[obsm_key][obsm_col]

    # Get data for plotting
    data = adata_obs[["axis", cell_type_column]].dropna()

    # Filter data for the specified cell type
    selected_data = data[data[cell_type_column] == cell_type].copy()

    # Apply limits
    selected_data = selected_data[
        (selected_data["axis"] >= xlim[0]) &
        (selected_data["axis"] <= xlim[1])
        ]

    for i, gene in enumerate(genes):
        # Add gene expression information
        gene_loc = adata.var_names.get_loc(gene)
        expr = adata.X[:, gene_loc]
        expr = pd.Series(expr.toarray().flatten(), index=adata.obs_names)

        if min_expression is not None:
            # mask values below the threshold with NaN
            expr = expr.mask(expr < min_expression)

        with warnings.catch_warnings():
            warnings.simplefilter(action="ignore", category=pd.errors.PerformanceWarning)
            # add gene expression to dataframe
            selected_data[gene] = expr

        # if min_expression is not None:
        #     # Apply limits
        #     #data_for_one_celltype = data_for_one_celltype[data_for_one_celltype[gene] >= min_expression]

    # drop the cell type column
    #selected_data = selected_data.drop(cell_type_column, axis=1)

    # add axis column to index
    selected_data = selected_data.set_index(["axis", cell_type_column], append=True)
    selected_data.index.names = ["cell_id", "axis", cell_type_column]

    if sort:
        #selected_data = selected_data.sort_values("axis")
        selected_data = selected_data.sort_index(level='axis', ascending=True)

    if minmax_scale:
        scaler = MinMaxScaler()
        selected_data = pd.DataFrame(scaler.fit_transform(selected_data),
                    index=selected_data.index, columns=selected_data.columns
                    )

    return selected_data

# binning function
def _bin_data(data,
              #expr, axis,
              axis_name: str = "axis",
              resolution=5,
              #minmax_scale: bool = True,
              plot: bool = False
              ):

    # make a copy of the data
    data = data.copy()

    # get values
    axis_data = data.index.get_level_values(axis_name).values

    # get gene names
    genes = [elem for elem in data.columns if elem != axis_name]

    if resolution > 0:
        # calculate number of bins from resolution
        nbins = int((axis_data.max() - axis_data.min()) / resolution)
        data["bin"] = pd.cut(axis_data, bins=nbins)

        # calculate mean per bin
        binned_mean = data.groupby("bin")[genes].mean()

        # extract the center of each bin
        binned_mean.index = [elem.mid for elem in binned_mean.index]
        binned_mean.index.name = "axis"
    else:
        # no actual binning is performed - data is just formated in the same way
        binned_mean = data.droplevel(["cell_id", "cell_type"])
        binned_mean.index.name = "axis"

    if plot:
        _bin_qc_plot(binned_data=binned_mean, raw_data=data)

    return binned_mean

def _bin_qc_plot(
    binned_data, raw_data, xlabel='x'
):
    # extract values from binned data
    bin_center = binned_data.index.values
    genes = binned_data.columns
    n_genes = len(genes)

    # extract values from raw data
    raw_axis = raw_data.index.get_level_values("axis").values

    fig, axs = plt.subplots(n_genes, 2, figsize=(8, 4*n_genes),
                            sharex='col', sharey='row'
                            )

    if n_genes == 1:
        # reshape the axis array so that 2d indexing works later
        axs = axs.reshape(1,2)

    for i, gene in enumerate(genes):
        expr = binned_data[gene].values
        raw_expr = raw_data[gene].values

        not_nan = ~np.isnan(expr)

        try:
            # perform loess regression for the second half of the plot
            res = smooth_fit(
            xs=bin_center[not_nan],
            ys=expr[not_nan], # make sure there are no NaN in the data
            loess_bootstrap=False, nsteps=100
            )
        except ValueError as e:
            print(f"A ValueError occurred during loess regression: {e}")
            res = None

        # Plot the original data
        axs[i, 0].scatter(
            raw_axis, raw_expr, label='Original Data', alpha=0.5, color='k', s=1
            )

        # Plot the binned values
        axs[i, 0].plot(
            bin_center, expr,
            color='firebrick',
            alpha=0.5,
            linestyle='-', label='Binned Mean')

        # Add labels and legend
        axs[i, 0].set_xlabel(xlabel)
        axs[i, 0].set_ylabel(f"Expression of '{gene}'")
        axs[i, 0].legend()

        #axs[1].plot(binned_data["bin_center"], binned_data["minmax"])
        axs[i, 1].scatter(
            x=bin_center, y=expr,s=1, color="k", label="Binned Mean")
        if res is not None:
            axs[i, 1].plot(res["x"], res["y_pred"], color='royalblue',
                           label="Loess Regression")
            axs[i, 1].fill_between(res["x"],
                                   res["conf_lower"],
                                   res["conf_upper"],
                                   color='royalblue',
                                   alpha=0.2, label='Confidence Interval')

        axs[i, 1].legend()
        axs[i, 1].set_xlabel(xlabel)
        axs[i, 1].set_ylabel(f"Scaled expression of '{gene}'")

    # Show plot
    plt.tight_layout()
    plt.show()

def cell_abundance_along_axis(
    adata: AnnData,
    axis: Union[str, Tuple[str, str]],
    groupby: Optional[str] = None,
    xlim: Tuple = (0, np.inf),
    savepath: Optional[os.PathLike] = None,
    figsize: Tuple = (8,6),
    save_only: bool = False,
    dpi_save: int = 300,
    multiple: Literal["layer", "dodge", "stack", "fill"] = "stack",
    histplot_element: Literal["bars", "step", "poly"] = "bars",
    kde: bool = False
    ):

    """
    Plot cell abundance along a specified observation value.

    Args:
        adata (AnnData): Annotated data matrix.
        axis (Union[str, Tuple[str, str]]): Observation category to be plotted on the x-axis.
            Can be a string representing a column in `adata.obs` or a tuple (obsm_key, obsm_col)
            where `obsm_key` is a key in `adata.obsm` and `obsm_col` is a column in the corresponding DataFrame.
        groupby (Optional[str], optional): Column in `adata.obs` to group by. Defaults to None.
        xmin (Number, optional): Minimum value of `obs_val` to include in the plot. Defaults to 0.
        savepath (Optional[os.PathLike], optional): Path to save the figure. Defaults to None.
        figsize (Tuple, optional): Size of the figure. Defaults to (8, 6).
        save_only (bool, optional): If True, only save the figure without displaying it. Defaults to False.
        dpi_save (int, optional): Dots per inch for saving the figure. Defaults to 300.
        histplot_multiple (str, optional): How to plot multiple histograms. Options are "layer", "dodge", "stack", "fill". Defaults to "stack".
        histplot_element (str, optional): Plotting element. Options are "bars", "step", "poly". Defaults to "bars".

    Returns:
        None
    """

    # check type of obs_val
    adata_obs = adata.obs.copy()
    if isinstance(axis, tuple):
        print("Retrieve `obs_val` from .obsm.")
        obsm_key = axis[0]
        obsm_col = axis[1]
        axis = f"distance_from_{obsm_col}"
        adata_obs[axis] = adata.obsm[obsm_key][obsm_col]

    # get data for plotting
    data = adata_obs[[axis, groupby]].dropna()

    # remove zeros
    xlim_mask = (data[axis] > xlim[0]) & (data[axis] <= xlim[1])
    data = data[xlim_mask].copy()

    # Create the histogram
    fig, ax = plt.subplots(1,1, figsize=(figsize[0], figsize[1]))

    if not kde:
        h = sns.histplot(data=data, x=axis,
                    hue=groupby, palette=DEFAULT_CATEGORICAL_CMAP.colors,
                    multiple=multiple, element=histplot_element,
                    alpha=1, ax=ax
                    )
    else:
        h = sns.kdeplot(data=data, x=axis,
                    hue=groupby, palette=DEFAULT_CATEGORICAL_CMAP.colors,
                    alpha=1, ax=ax, multiple=multiple
                    )
        plt.xlim(0, data[axis].max())

    # Move the legend outside of the plot
    sns.move_legend(h, "upper left", bbox_to_anchor=(1, 1))

    # save or show figure
    save_and_show_figure(savepath=savepath,
                         fig=h.get_figure(),
                         save_only=save_only,
                         dpi_save=dpi_save,
                         tight=True
                         )

# def cell_expression_along_axis(
#     adata,
#     obs_val,
#     genes,
#     cell_type_column,
#     cell_type,
#     xlim: Tuple[Number, Number] = (0, np.inf),
#     min_expression: Number = 0,
#     xlabel: Optional[str] = None,
#     fit_reg: bool = False,
#     lowess: bool = False,
#     robust: bool = False
#     ):
#     genes = convert_to_list(genes)

#     # if len(genes) == 1:
#     #     _single_cell_expression_along_axis(
#     #         adata=adata, obs_val=obs_val, gene=genes[0],
#     #         cell_type_column=cell_type_column, cell_type=cell_type,
#     #         xlim=xlim, min_expression=min_expression, xlabel=xlabel,
#     #         fit_reg=fit_reg, lowess=lowess, robust=robust
#     #     )
#     # elif len(genes) > 1:
#     _multi_cell_expression_along_axis(
#         adata=adata, obs_val=obs_val, genes=genes,
#         cell_type_column=cell_type_column, cell_type=cell_type,
#         xlim=xlim, min_expression=min_expression, xlabel=xlabel,
#         fit_reg=fit_reg, lowess=lowess, robust=robust
#     )
#     # else:
#     #     raise ValueError("`genes` must have length > 0.")

# def _single_cell_expression_along_axis(
#     adata,
#     obs_val,
#     gene,
#     cell_type_column,
#     cell_type,
#     xlim: Tuple[Number, Number] = (0, np.inf),
#     min_expression: Number = 0,
#     xlabel: Optional[str] = None,
#     fit_reg: bool = False,
#     lowess: bool = False,
#     robust: bool = False
#     ):

#     data_of_one_celltype = _select_data(
#         adata=adata,
#         obs_val=obs_val,
#         cell_type_column=cell_type_column,
#         cell_type=cell_type,
#         genes=gene,
#         xlim=xlim,
#     )

#     # Filter for minimum gene expression
#     data_of_one_celltype = data_of_one_celltype[data_of_one_celltype[gene] >= min_expression]

#     # Plot
#     g = sns.jointplot(data=data_of_one_celltype,
#                     x="axis", y=gene,
#                     height=4,
#                     color="firebrick", kind="kde", levels=8,
#                     marginal_kws={"fill": True},
#                     )
#     #g.plot_joint(sns.scatterplot, color="k", s=12)
#     g.plot_joint(sns.regplot, color="k",
#                  #lowess=True,
#                  fit_reg=fit_reg,
#                  lowess=lowess,
#                  robust=robust,
#                  scatter_kws={"s": 1},
#                  line_kws={"color": "orange"}
#                  )
#     g.ax_joint.set_ylabel(f"{gene} in '{cell_type}'")

#     # Set common x-label
#     if xlabel is None:
#         g.ax_joint.set_xlabel("_".join(convert_to_list(obs_val)))
#     else:
#         g.ax_joint.set_xlabel(xlabel)

#     # g = sns.jointplot(data=data_of_one_celltype,
#     #             x=axis_label, y=gene,
#     #             color="k", kind="reg", #levels=8,
#     #             marginal_kws={"fill": True},
#     #             )
#     # g.plot_marginals(sns.kdeplot, color="firebrick", #s=12
#     #                  )
#     plt.show()


from typing import List, Optional, Union

import numpy as np
import pandas as pd
from anndata._core.anndata import AnnData

from insitupy.utils.utils import convert_to_list


def create_deg_dataframe(
    adata: AnnData,
    groups: Optional[str] = None,
    key: str = 'rank_genes_groups'
    ) -> dict:
    """
    Create a DataFrame of differentially expressed genes from rank_genes_groups results.
    Can be used to plot a volcano plot using `plot_volcano`.

    This function extracts gene names, log fold changes, and p-values from the
    specified group in the rank_genes_groups results of an AnnData object.
    It also handles zero p-values to avoid log10(0) errors.

    Args:
        adata (AnnData): The AnnData object containing the results.
        groups (str or None): The name of the group to extract results for.
                                   If None, create DataFrames for all groups.
        key (str): The key in adata.uns where the rank_genes_groups results are stored.
                   Defaults to 'rank_genes_groups'.

    Returns:
        dict:
            - Returns a dictionary of DataFrames for all groups,
              where each key is a group name and the value is the corresponding DataFrame.
    """
    results = adata.uns[key]

    if groups is not None:
        groups = convert_to_list(groups)
    else:
        groups = results['names'].dtype.names

    # Create a dictionary of DataFrames for all groups
    volcano_data_dict = {}

    for group in groups:
        volcano_data = pd.DataFrame({
            'gene': results['names'][group],
            'logfoldchanges': results['logfoldchanges'][group],
            'pvals': results['pvals'][group],
            'scores': results['scores'][group],
        })
        # Replace zero p-values with a small value to avoid log10(0)
        volcano_data['pvals'] = volcano_data['pvals'].replace(0, 1e-300)
        volcano_data['neg_log10_pvals'] = -np.log10(volcano_data['pvals'])
        volcano_data_dict[group] = volcano_data

    return volcano_data_dict




import os
from pathlib import Path

import matplotlib.pyplot as plt


def save_and_show_figure(savepath, fig, save_only=False, show=True, dpi_save=300, background_color=None, tight=True):

    if tight:
        fig.tight_layout()

    if savepath is not None:
        print("Saving figure to file " + savepath)

        # create path if it does not exist
        Path(os.path.dirname(savepath)).mkdir(parents=True, exist_ok=True)

        # save figure
        plt.savefig(savepath, dpi=dpi_save,
                    facecolor=background_color, bbox_inches='tight')
        print("Saved.")
    if save_only:
        plt.close(fig)
    elif show:
        return plt.show()
    else:
        return


import textwrap

import matplotlib.pyplot as plt
import numpy as np

from insitupy.io.plots import save_and_show_figure
from insitupy.utils.utils import check_list, get_nrows_maxcols


def go_plot(
    enrichment,
    style="dot",
    color_key="Gene ratio",
    size_key=None, groups=None,
    fig=None, axis=None,
    max_to_plot=10,
    max_cols=4, cmap='viridis', cmin=None, cmax=None,
    ax_label_size=16, markersize=240,
    figsize=(8,6),
    remove_yticklabels=False,
    xtick_label_size=16, ytick_label_size=16,
    clb_label_size=16, clb_tick_label_size=16,
    clb_pos=None, clb_norm=False,
    title_size=16, max_line_length=30, custom_headers=None,
    max_name_length=60,
    value_to_plot='Enrichment score',
    colorcol=None,
    sortby=None, sort=True, additional_sortby=None,
    x_margin=None, y_margin=None,
    name_key='name', libraries=None, ascending=False,
    savepath=None, save_only=False, show=True, dpi_save=300
    ):

    # get groups from index of enrichment dataframe
    extracted_groups = enrichment.index.unique(level=0).tolist()

    if groups is not None:
        groups = [groups] if isinstance(groups, str) else list(groups)

        groups = check_list(groups, extracted_groups)

        if len(groups) == 0:
            return
    else:
        groups = extracted_groups

    # check possible custom headers
    if custom_headers is not None:
        custom_headers = [custom_headers] if isinstance(custom_headers, str) else list(custom_headers)

    # check for libraries
    if libraries is None:
        libraries = list(enrichment['source'].unique())
    else:
        # convert libraries to list if they are string
        libraries = [libraries] if isinstance(libraries, str) else list(libraries)

        # check if all libraries are in dataframe
        libs_available = enrichment['source'].unique()
        notin = [elem for elem in libraries if elem not in libs_available]
        assert len(notin) == 0, (
            f"Following libraries could not be found in the `source` column: {notin}\n"
            f"Following libraries are possible: {libs_available}"
        )
        # filter for libraries
        enrichment = enrichment[enrichment['source'].isin(libraries)].copy()

    # sort dataframe
    if sortby is None:
        sortby = value_to_plot

    if sort:
        enrichment.sort_values(sortby, ascending=ascending, inplace=True)

    # shorten to max length if necessary
    if max_to_plot is not None:
        enrichment = enrichment.groupby(level=0).head(max_to_plot).copy()

    if additional_sortby is not None:
        enrichment.sort_values([additional_sortby, sortby], ascending=[True, False], inplace=True)
        #enrichment[name_key] = ["{} ({})".format(a,b) for a,b in zip(enrichment[name_key], enrichment[additional_sortby])]

    # Prepare names for plotting
    # Shorten name and add GO term name if too short
    if max_name_length is not None:
        enrichment[name_key] = ["{}...({})".format(n[:max_name_length], go) if len(n)>max_name_length else n for go, n in zip(enrichment['native'], enrichment[name_key])]
    # Insert line breaks if sentence too long
    #enrichment[name_key] = [nth_repl_all(elem, ' ', '\n', 1) for elem in enrichment[name_key]]

    # get min and max for the colorbar
    if color_key is not None:
        try:
            if clb_norm:
                cmax = enrichment[color_key].max() if cmax is None else cmax
                cmin = enrichment[color_key].min() if cmin is None else cmin
            else:
                cmax = None
                cmin = None
        except KeyError:
            print(f"color_key '{color_key}' not found in columns.")

    # Plotting
    if axis is None:
        n_plots, n_rows, n_cols = get_nrows_maxcols(len(groups), max_cols=max_cols)
        fig, axs = plt.subplots(n_rows, n_cols, figsize=(figsize[0]*n_cols, figsize[1]*n_rows))

    else:
        assert len(groups) == 1, "Enrichment dataframe contains more than one group in first index level 0."
        axs = axis
        n_plots = 1
        show = False

    axs = axs.ravel() if n_plots > 1 else [axs]

    for i, group in enumerate(groups):
        if group in enrichment.index.unique(level=0):
            # select data
            df = enrichment.xs(group).copy()

            # if max_to_plot is not None and len(df) > max_to_plot:
            #         df = df[:max_to_plot]
            if color_key is not None:
                try:
                    color = df[color_key]
                except KeyError:
                    print(f"color_key '{color_key}' not found in columns.")
                    color = 'k'
            else:
                color = 'k'

            if size_key is not None:
                markersize=df[size_key]*5

            if max_line_length is not None:
                # introduce line breaks if the names are too long
                df[name_key] = [textwrap.fill(elem, width=max_line_length, break_long_words=True) for elem in df[name_key]]

            # plotting
            if style == "dot":
                s = axs[i].scatter(
                    df[value_to_plot], df[name_key],
                    c=color, cmap=cmap,
                    s=markersize,
                    edgecolors='k')
            elif style == "dot_fixed":
                s = axs[i].scatter(
                    x = [0] * len(df[name_key]),
                    y = df[name_key],
                    s=df[value_to_plot] * 50,
                    c=color,
                    cmap=cmap,
                    edgecolors='k')
            elif style == "bar":
                ys = df[name_key]
                ys_pos = np.arange(len(ys))
                s = axs[i].barh(
                    ys_pos,
                    width=df[value_to_plot],
                    height=0.8,
                    color='k',
                    #cmap=cmap,
                    #s=markersize,
                    #edgecolors='k'
                )
                axs[i].set_yticks(ys_pos, labels=ys)
            else:
                raise ValueError('Invalid `style` parameter ("{}"). Possible options'.format(style))

            if custom_headers is None:
                axs[i].set_title("{}\n{}".format(group, libraries), fontsize=title_size)
            else:
                axs[i].set_title(custom_headers[i], fontsize=title_size)

            axs[i].invert_yaxis()
            axs[i].set_xlabel(value_to_plot, fontsize=ax_label_size)
            axs[i].tick_params(axis='x', which='major', labelsize=xtick_label_size)
            axs[i].tick_params(axis='y', which='major', labelsize=ytick_label_size)
            axs[i].margins(x=x_margin, y=y_margin)

            if style == "dot":
                axs[i].grid(axis='y')

            if style == "dot_fixed":
                axs[i].spines['top'].set_visible(False)
                axs[i].spines['right'].set_visible(False)
                axs[i].spines['bottom'].set_visible(False)
                axs[i].spines['left'].set_visible(False)
                axs[i].get_xaxis().set_ticks([])
                axs[i].set_xlabel("")

            if remove_yticklabels:
                axs[i].set_yticklabels([])

            # cbar_ax = fig.add_axes([0.95, 0.12, 0.02, 0.75])
            # clb = plt.colorbar(s, cax=cbar_ax, orientation='vertical')
            # clb.set_label('Fraction of patterned genes', loc='center')

            if style in ["dot", "dot_fixed"]:
                if color_key is not None:
                    if clb_pos is None:
                        clb = fig.colorbar(s, ax=axs[i], fraction=0.2, aspect=40, pad=0.15)
                        clb.set_label(color_key, loc='center', fontsize=clb_label_size)
                        clb.ax.tick_params(labelsize=clb_tick_label_size)
                        if clb_norm:
                            clb.mappable.set_clim(cmin, cmax)
                    else:
                        if i == clb_pos:
                            clb = fig.colorbar(s, ax=axs[clb_pos])
                            clb.set_label(color_key, loc='center', fontsize=clb_label_size)
                            clb.ax.tick_params(labelsize=clb_tick_label_size)
                            if clb_norm:
                                clb.mappable.set_clim(cmin, cmax)



            if size_key is not None:
                kw = dict(prop="sizes", num=5,
                    #color=s.cmap(0.7),
                    #fmt="$ {x:.2f}",
                    #func=lambda s: np.sqrt(s/.3)/3
                    #func=lambda s: np.sqrt(s)
                    )
                size_legend = axs[i].legend(*s.legend_elements(**kw, alpha=0.6),
                                    #markerscale=0.5,
                                    loc="lower right", title=size_key)

            if colorcol is not None:
                color_dict = {a: b for a, b in zip(df["name"], df[colorcol])}
                for xtick in axs[i].get_yticklabels():
                    xtick.set_color(color_dict[xtick.get_text()])
                # for xtick, color in zip(axs[i].get_xticklabels(), enrichment[colorcol]):
                #     xtick.set_color(color)


        else:
            #print('No significant results for selected libraries {} in group {}'.format(libraries, group))
            axs[i].set_title("{}\n{}".format(group, libraries), fontsize=12, fontweight='bold')
            axs[i].text(0.5,0.5, 'No significant results for selected\nlibraries {}\nin group {}'.format(libraries, group),
                        va='center', ha='center', fontsize=20)
            axs[i].set_axis_off()

    # check if there are empty plots remaining
    if n_plots > 1:
        while i < n_rows * n_cols - 1:
            i += 1
            # remove empty plots
            axs[i].set_axis_off()

        fig.tight_layout()

    fig.tight_layout()
    if show:
        save_and_show_figure(savepath=savepath, save_only=save_only, dpi_save=dpi_save, fig=fig)
    else:
        return fig, axs


import logging
import sys
from numbers import Number
from typing import List, Literal, Optional, Tuple, Union
from warnings import warn

import matplotlib
import numpy as np
import pandas as pd
from matplotlib.colors import rgb2hex
from pandas.api.types import is_numeric_dtype
from shapely import (LinearRing, LineString, MultiPoint, MultiPolygon, Point,
                     Polygon)

from insitupy import WITH_NAPARI
from insitupy._constants import (ANNOTATIONS_SYMBOL, DEFAULT_CATEGORICAL_CMAP,
                                 DEFAULT_CONTINUOUS_CMAP, POINTS_SYMBOL,
                                 REGION_CMAP, REGIONS_SYMBOL)
from insitupy.palettes import CustomPalettes
from insitupy.plotting._colors import _data_to_rgba, _determine_climits

from ._checks import check_rgb_column

if WITH_NAPARI:
    import napari
    from napari.types import LayerDataTuple

    def _add_geometries_as_layer(
        dataframe: pd.DataFrame,
        viewer: napari.Viewer,
        layer_name: str,
        #scale_factor: Union[Tuple, List, np.ndarray],
        edge_width: Number = 10, # µm
        opacity: float = 1,
        rgb_color: Optional[Tuple] = None,
        show_names: bool = False,
        allow_duplicate_layers: bool = False,
        mode: Literal["Annotations", "Regions"] = "Annotations",
        tolerance: Number = 5
        ):

        # list to store information on shapes
        shape_list = []
        shape_type_list = []

        # lists to store the point x and y coordinates
        point_x_list = []
        point_y_list = []

        # iterate through annotations of this class and collect them as list
        color_list = {"Points": [], "Shapes": []}
        uid_list = {"Points": [], "Shapes": []}
        type_list = {"Points": [], "Shapes": []} # list to store whether the polygon is exterior or interior
        names_list = {"Points": [], "Shapes": []}

        # check if colors are given
        if "color" in dataframe.columns:
            # make sure the RGB column consists only of valid RGB tuples or lists
            rgbs_valid = check_rgb_column(dataframe, "color")
            if not rgbs_valid:
                warn('Not all RGB values given in column "color" are valid. Used blue for all geometries.')
                dataframe["color"] = [(0,0,255)] * len(dataframe)
        else:
            if rgb_color is not None:
                dataframe["color"] = [rgb_color] * len(dataframe)
            else:
                dataframe["color"] = [(0,0,255)] * len(dataframe)

        for uid, row in dataframe.iterrows():
            # get coordinates
            geometry = row["geometry"]
            #uid = row["id"]
            hexcolor = rgb2hex([elem / 255 for elem in row["color"]])

            # check if polygon is a MultiPolygon or just a simple Polygon object
            if isinstance(geometry, MultiPolygon):
                data = list(geometry.geoms)
                annotation_type = "polygon_like"
            elif isinstance(geometry, Polygon):
                data = [geometry]
                annotation_type = "polygon_like"
            elif isinstance(geometry, LineString):
                data = geometry
                annotation_type = "line_like"
            elif isinstance(geometry, Point) or isinstance(geometry, MultiPoint):
                data = geometry
                annotation_type = "point_like"
            else:
                raise ValueError(f"Received unknown geometry type: {type(geometry)}")

            if annotation_type == "polygon_like":
                for p in data:
                    # simplify polygon for visualization
                    p = p.simplify(tolerance)
                    # extract exterior coordinates from shapely object
                    # Note: the last coordinate is removed since it is identical with the first
                    # in shapely objects, leading sometimes to visualization bugs in napari
                    exterior_array = np.array([p.exterior.coords.xy[1].tolist()[:-1],
                                            p.exterior.coords.xy[0].tolist()[:-1]]).T
                    shape_list.append(exterior_array)  # collect shape
                    color_list["Shapes"].append(hexcolor)  # collect corresponding color
                    uid_list["Shapes"].append(uid)  # collect corresponding unique id
                    type_list["Shapes"].append("polygon_exterior")
                    names_list["Shapes"].append(row["name"])
                    shape_type_list.append("polygon")

                    # if polygon has interiors, plot them as well
                    # for information on donut-shaped polygons in napari see: https://forum.image.sc/t/is-it-possible-to-generate-doughnut-shapes-in-napari-shapes-layer/88834
                    if len(p.interiors) > 0:
                        for linear_ring in p.interiors:
                            if isinstance(linear_ring, LinearRing):
                                interior_array = np.array([linear_ring.coords.xy[1].tolist()[:-1],
                                                        linear_ring.coords.xy[0].tolist()[:-1]]).T
                                shape_list.append(interior_array)  # collect shape
                                color_list["Shapes"].append(hexcolor)  # collect corresponding color
                                uid_list["Shapes"].append(uid)  # collect corresponding unique id
                                type_list["Shapes"].append("polygon_interior")
                                names_list["Shapes"].append(row["name"])
                                shape_type_list.append("polygon")
                            else:
                                ValueError(f"Input must be a LinearRing object. Received: {type(linear_ring)}")

            elif annotation_type == "line_like":
                line_array = np.array([data.coords.xy[1].tolist(), data.coords.xy[0].tolist()]).T

                # collect data
                shape_list.append(line_array)
                color_list["Shapes"].append(hexcolor)  # collect corresponding color
                uid_list["Shapes"].append(uid)  # collect corresponding unique id
                type_list["Shapes"].append("line") # information on type of coordinates - important for interior/exterior of polygons
                names_list["Shapes"].append(row["name"])
                shape_type_list.append("path")

            elif annotation_type == "point_like":
                try:
                    # in case of MultiPoints we first have to extract the individual geometries
                    point_coords = [elem.coords.xy for elem in geometry.geoms]
                except AttributeError:
                    # a normal Point object does not have multiple geometries and coordinates can be accessed directly
                    point_coords = [geometry.coords.xy]

                # collect coordinates and other data on the points
                for coord in point_coords:
                    point_x_list.append(coord[1].tolist()[0])
                    point_y_list.append(coord[0].tolist()[0])
                    color_list["Points"].append(hexcolor)  # collect corresponding color
                    uid_list["Points"].append(uid)  # collect corresponding unique id
                    type_list["Points"].append("point") # information on type of coordinates - important for interior/exterior of polygons
                    names_list["Points"].append(row["name"])

        if len(shape_list) > 0:
            properties_dict = {
                    'uid': uid_list["Shapes"], # list with uids
                    'type': type_list["Shapes"] # list giving information on whether the polygon is interior or exterior
                }
            if show_names:
                properties_dict['name'] = names_list["Shapes"]

                text_dict = {
                    'string': '{name}',
                    'anchor': 'upper_left',
                    #'translation': [-5, 0],
                    'size': 8,
                    'color': color_list["Shapes"]
                    }
            else:
                text_dict = None

            if mode == "Regions":
                layer_name_with_symbol = REGIONS_SYMBOL + " " + layer_name
            elif mode == "Annotations":
                layer_name_with_symbol = ANNOTATIONS_SYMBOL + " " + layer_name
            else:
                raise ValueError(f"Unknown value for `mode`: {mode}")

            # add shapes to viewer
            if not layer_name_with_symbol in viewer.layers:
                add = True
            elif allow_duplicate_layers:
                add = True
            else:
                add = False

            if add:
                viewer.add_shapes(
                    data=shape_list,
                    name=layer_name_with_symbol,
                    properties=properties_dict,
                    shape_type=shape_type_list,
                    edge_width=edge_width, # µm
                    edge_color=color_list["Shapes"],
                    face_color='transparent',
                    opacity=opacity,
                    #scale=scale_factor,
                    text=text_dict
                    )

        point_data = np.stack([point_x_list, point_y_list]).T
        if len(point_data) > 0:
            layer_name_with_symbol = POINTS_SYMBOL + " " + layer_name
            # add points to viewer
            # add shapes to viewer
            if not layer_name_with_symbol in viewer.layers:
                add = True
            elif allow_duplicate_layers:
                add = True
            else:
                add = False

            if add:
                viewer.add_points(
                    data=point_data,
                    name=layer_name_with_symbol,
                    properties={
                        'uid': uid_list["Points"], # list with uids
                        'type': type_list["Points"] # list giving information on whether the polygon is interior or exterior
                    },
                    size=40,
                    edge_color="black",
                    face_color=color_list["Points"],
                    #scale=scale_factor
                )


    def _create_points_layer(points,
                            color_values: List[Number],
                            name: str,
                            point_names: List[str],
                            point_size: int = 6, # is in scale unit (so mostly µm)
                            opacity: float = 1,
                            visible: bool = True,
                            edge_width: float = 0,
                            edge_color: str = 'red',
                            upper_climit_pct: int = 99,
                            continuous_cmap = DEFAULT_CONTINUOUS_CMAP,
                            categorical_cmap = DEFAULT_CATEGORICAL_CMAP
                            ) -> LayerDataTuple:

        # get colors
        colors, mapping, cmap = _data_to_rgba(data=color_values,
                               continuous_cmap=continuous_cmap,
                               categorical_cmap=categorical_cmap,
                               upper_climit_pct=upper_climit_pct)

        # generate point layer
        layer = (
            points,
            {
                'name': name,
                'properties': {
                    "value": color_values,
                    "cell_name": point_names
                    },
                'symbol': 'o',
                'size': point_size,
                'face_color': colors,
                'opacity': opacity,
                'visible': visible,
                'edge_width': edge_width,
                'edge_color': edge_color,
                'metadata': {"upper_climit_pct": upper_climit_pct}
                },
            'points'
            )
        return layer

    def _update_points_layer(
        layer: napari.layers.Layer,
        new_color_values: List[Number],
        new_name: Optional[str] = None,
        upper_climit_pct: int = 99,
        # cmap: str = "viridis"
        ) -> None:
        # get the RGBA colors for the new values
        new_colors, mapping, cmap = _data_to_rgba(data=new_color_values, upper_climit_pct=upper_climit_pct)

        # change the colors of the layer
        layer.face_color = new_colors

        # change properties of layer
        new_props = layer.properties.copy()
        new_props['value'] = new_color_values
        layer.properties = new_props

        if new_name is not None:
            layer.name = new_name

#TODO: Why is this function not used anywhere?
def _determine_color_settings(
    color_values,
    cmap,
    upper_climit_pct
    ):
    # check if the data should be plotted categorical or continous
    if is_numeric_dtype(color_values):
        is_categorical = False # if the data is numeric it should be plotted continous
    else:
        is_categorical = True # if the data is not numeric it should be plotted categorically

    if is_categorical:
        # get color cycle for categorical data
        color_mode = "cycle"
        # palettes = CustomPalettes()
        # color_cycle = getattr(palettes, "tab20_mod").colors
        color_cycle = DEFAULT_CATEGORICAL_CMAP.colors
        color_map = None
        climits = None
    else:
        color_mode = "colormap"
        color_map = cmap
        color_cycle = None

        climits = _determine_climits(
            color_values=color_values,
            upper_climit_pct=upper_climit_pct
        )

    return color_mode, color_cycle, color_map, climits


import os
from pathlib import Path

import anndata
import pandas as pd


class XeniumPanels:
    '''
    Class containing a collection of custom color palettes.
    '''
    def __init__(self, verbose=False):
        # read all panels
        script_dir = Path(os.path.realpath(__file__)).parent
        panel_dir = script_dir / Path("../xenium_panels/")
        panel_dir = panel_dir.resolve()
        panel_paths = sorted(panel_dir.glob("*.csv"))

        for p in panel_paths:
            name = p.stem.split("_")[1]
            panel = pd.read_csv(p)
            setattr(self, name, panel)
            print(name) if verbose else None

            # make sure that the column names are correct
            panel.columns = ["Gene", "Ensembl_ID", "Coverage", "Codewords", "Annotation"]
        
    def show_all(self):
        '''
        Prints all available panels.
        '''
        panel_list = [elem for elem in dir(self) if not elem.startswith("__")]
        panel_list = [elem for elem in panel_list if elem not in ["show_all"]]
        for p in panel_list:
            print(p)
            
            
def generate_mock_reference(
    dataframe: pd.DataFrame,
    annotation_column: str = "Annotation",
    gene_column: str = "Gene"
    ):

    # Initialize an empty dictionary to store marker genes for each annotation
    annotation_genes = {}

    # Iterate over unique annotations
    for annotation in dataframe[annotation_column].unique():
        # Collect marker genes for the annotation
        marker_genes = dataframe.loc[dataframe[annotation_column] == annotation, gene_column].tolist()
        # Add the annotation and its marker genes to the dictionary
        annotation_genes[annotation] = marker_genes

    # Create a DataFrame to store the count matrix
    count_matrix = pd.DataFrame(0, columns=dataframe[gene_column], index=dataframe[annotation_column].unique())

    # Set marker genes to 1000 in the count matrix
    for annotation, genes in annotation_genes.items():
        count_matrix.loc[annotation, genes] = 1000

    # Create annotation DataFrame
    annotation_df = pd.DataFrame({'Annotation': count_matrix.index})

    # match index
    annotation_df.index = count_matrix.index
    # create adata
    reference = anndata.AnnData(X = count_matrix, obs = annotation_df)
    
    return reference


import os
from pathlib import Path
from typing import Union

import geopandas as gpd
import shapely

from insitupy.io.files import read_json


def read_baysor_polygons(
    file: Union[str, os.PathLike, Path]
    ) -> gpd.GeoDataFrame:

    d = read_json(file)

    # prepare output dictionary
    df = {
    "geometry": [],
    "cell": [],
    "type": [],
    "minx": [],
    "miny": [],
    "maxx": [],
    "maxy": []
    }

    for elem in d["geometries"]:
        coords = elem["coordinates"][0]

        # check if there are enough coordinates for a Polygon (some segmented cells are very small in Baysor)
        if len(coords) > 3:
            p = shapely.Polygon(coords)
            df["geometry"].append(p)
            df["type"].append("polygon")

        else:
            p = shapely.LineString(coords)
            df["geometry"].append(p)
            df["type"].append("line")
        df["cell"].append(elem["cell"])

        # extract bounding box
        bounds = p.bounds
        df["minx"].append(bounds[0])
        df["miny"].append(bounds[1])
        df["maxx"].append(bounds[2])
        df["maxy"].append(bounds[3])

    # create geopandas dataframe
    df = gpd.GeoDataFrame(df)

    return df


import os
import warnings
from numbers import Number
from pathlib import Path
from typing import List, Literal, Union

import dask.array as da
import numpy as np
import pandas as pd
import scanpy as sc
import zarr
from anndata import AnnData
from pandas.api.types import is_numeric_dtype
from scipy.sparse import csr_matrix
from zarr.errors import ArrayNotFoundError

from insitupy._core.dataclasses import BoundariesData
from insitupy._exceptions import InvalidFileTypeError
from insitupy.utils.utils import (convert_int_to_xenium_hex,
                                  decode_robust_series)


def _read_matrix_from_xenium(path) -> AnnData:
    # extract parameters from metadata
    path = Path(path)
    cf_h5_path = path / "cell_feature_matrix.h5"

    with warnings.catch_warnings():
        warnings.simplefilter(action='ignore', category=FutureWarning)
        # read matrix data
        adata = sc.read_10x_h5(cf_h5_path)

    # read cell information
    cells_parquet_path = path / "cells.parquet"
    cells = pd.read_parquet(cells_parquet_path)

    # transform cell ids from bytes to str
    cells = cells.set_index("cell_id")

    # make sure that the indices are decoded strings
    if is_numeric_dtype(cells.index):
        cells.index = cells.index.astype(str)
    else:
        cells.index = decode_robust_series(cells.index)

    # add information to anndata observations
    adata.obs = pd.merge(left=adata.obs, right=cells, left_index=True, right_index=True)

    # transfer coordinates to .obsm
    coord_cols = ["x_centroid", "y_centroid"]
    adata.obsm["spatial"] = adata.obs[coord_cols].values
    adata.obsm["spatial"]
    adata.obs.drop(coord_cols, axis=1, inplace=True)

    return adata


def _read_boundaries_from_xenium(
    path: Union[str, os.PathLike, Path],
    pixel_size: Number,
    mode: Literal["dataframe", "mask"] = "mask"
    ) -> BoundariesData:
    # # read boundaries data
    path = Path(path)

    if mode == "dataframe":
        files=["cell_boundaries.parquet", "nucleus_boundaries.parquet"]
        labels=["cellular", "nuclear"]

        # generate path for files
        files = [path / f for f in files]

        # generate dataframes
        data_dict = {}
        for n, f in zip(labels, files):
            # check the file suffix
            if not f.suffix == ".parquet":
                InvalidFileTypeError(allowed_types=[".parquet"], received_type=f.suffix)

            # load dataframe
            df = pd.read_parquet(f)

            # decode columns
            df = df.apply(lambda x: decode_robust_series(x), axis=0)

            # collect dataframe
            data_dict[n] = df

        # create boundariesdata object
        boundaries = BoundariesData()

    else:
        cells_zarr_file = path / "cells.zarr.zip"

        # open zarr directory using dask
        data_dict = {
            "cells": da.from_zarr(cells_zarr_file, component="masks/1"),
            "nuclei": da.from_zarr(cells_zarr_file, component="masks/0")
        }

        # read cell ids and seg mask value
        # for info see: https://www.10xgenomics.com/support/software/xenium-onboard-analysis/latest/analysis/xoa-output-zarr#cells
        cell_ids = da.from_zarr(cells_zarr_file, component="cell_id").compute()
        if len(cell_ids.shape) == 2:
            cell_names = np.array([convert_int_to_xenium_hex(elem[0], elem[1]) for elem in cell_ids])
        elif len(cell_ids.shape) == 1:
            cell_names = cell_ids.astype(str)
        else:
            raise ValueError(f"Unexpected shape for `cell_ids` array: {cell_ids.shape} instead of 1 or 2.")

        try:
            seg_mask_value = da.from_zarr(cells_zarr_file, component="seg_mask_value")
        except ArrayNotFoundError:
            seg_mask_value = np.array(range(1, len(cell_names)+1))
            #seg_mask_value = None

        # create boundariesdata object
        boundaries = BoundariesData(
            #cell_ids=cell_ids,
            cell_names=cell_names,
            seg_mask_value=seg_mask_value
            )

    boundaries.add_boundaries(data=data_dict,
                              pixel_size=pixel_size)

    return boundaries


def _read_binned_expression(
    path: Union[str, os.PathLike, Path],
    gene_names_to_select = List
):
    path = Path(path)
    # add binned expression data to .varm of self.cells.matrix
    trans_file = path / "transcripts.zarr.zip"

    # read zarr store
    t = zarr.open(trans_file, mode="r")

    # extract sparse array
    data_gene = t["density/gene"]
    data = data_gene["data"][:]
    indices = data_gene["indices"][:]
    indptr = data_gene["indptr"][:]

    # get dimensions of the array
    cols = data_gene.attrs["cols"]
    rows = data_gene.attrs["rows"]

    # get info on gene names
    gene_names = data_gene.attrs["gene_names"]
    n_genes = len(gene_names)

    sarr = csr_matrix((data, indices, indptr))

    # reshape to get binned data
    arr = sarr.toarray()
    arr = arr.reshape((n_genes, rows, cols))

    # select only genes that are available in the adata object
    gene_mask = [elem in gene_names_to_select for elem in gene_names]
    arr = arr[gene_mask]
    return arr


def _restructure_transcripts_dataframe(dataframe: pd.DataFrame, decode:bool = False):

    if decode:
        # decode columns
        dataframe = dataframe.apply(lambda x: decode_robust_series(x), axis=0)

    # set index and rename columns
    dataframe = dataframe.set_index("transcript_id")
    dataframe = dataframe.rename({
        "cell_id": "cell_id",
        "x_location": "x",
        "y_location": "y",
        "z_location": "z",
        "feature_name": "gene"
    }, axis=1)

    # reorder dataframe
    column_names_ordered = ["x", "y", "z", "gene", "qv", "overlaps_nucleus", "fov_name", "nucleus_distance", "cell_id"]
    in_df = [elem in dataframe.columns for elem in column_names_ordered]
    column_names_ordered = [elem for i, elem in zip(in_df, column_names_ordered) if i]
    dataframe = dataframe.loc[:, column_names_ordered]

    # group column names into MultiIndices
    grouped_column_names = [
        ("coordinates", "x"),
        ("coordinates", "y"),
        ("coordinates", "z"),
        ("properties", "gene"),
        ("properties", "qv"),
        ("properties", "overlaps_nucleus"),
        ("properties", "fov_name"),
        ("properties", "nucleus_distance"),
        ("cell_id", "xenium")
    ]
    grouped_column_names = [elem for i, elem in zip(in_df, grouped_column_names) if i]
    dataframe.columns = pd.MultiIndex.from_tuples(grouped_column_names)
    return dataframe